// ##################################################################################################
// Program to calculate equinoxes, solstices, new moons, and full moons for given year.
// Program employs VSOP87 model of planetary motion, and ELP2000-82B model of lunar motion.
// ##################################################################################################

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "Utilities/Constants.h"
#include "Utilities/Time.h"
#include "Utilities/Frames.h"
#include "Utilities/VSOP_ELP.h"
#include "Utilities/Ecliptic.h"

// ###################
// Function prototypes
// ###################

// Function to determine equinoxes and solstices of given year
void Equinox (int Y);

// Function to determine dates of all full moons in given year
void FullMoon (int Y);

// Function to determine dates of all new moons in given year.
void NewMoon (int Y);

// Function to return geocentric elongation of Moon from Sun.
double MoonElongation (double t);

// #############
// Main function
//##############

int main ()
{
  // Output welcome message
  printf ("\n");
  printf ("========================================================================================\n");
  printf ("Program to calculate equinoxes, solstices, new moons, and full moons for given year.\n");
  printf ("Program employs VSOP87 model of planetary motion, and ELP2000-82B model of lunar motion.\n");
  printf ("========================================================================================\n");

  // Select year
  printf ("\n");
  printf ("Select year (YYYY) ");
  int Y;
  if (scanf ("%d", &Y) != 1)
    {
      printf ("Fatal Error: Cannot read year number\n");
      exit (1);
    }

  // Determine dates of equinoxes and solstices
  Equinox (Y);

  // Determine dates of full moons
  FullMoon (Y);

  // Determine dates of new moons
  NewMoon (Y);
}

// ####################
// Function definitions
// ####################

// ===========================================================
// Function to determine equinoxes and solstices of given year
// Y is year (e.g., 1963).
// ===========================================================
void Equinox (int Y)
{
  printf ("\n");
  
  // Find spring equinox
  int    M = 3;
  double D = 21.;
  double JD, lambda;
  DatetoJulian (Y, M, D, JD);
  do
    {
      // Get geocentric ecliptic coordinates of Sun
      double r, beta;
      SunEcliptic (JD, r, beta, lambda, 1, 1, 1, 1);

      JD -= (lambda - 0.) * (365.24/2./M_PI);
    }
  while (fabs (lambda) > 1.e-10);

  JuliantoDate (JD, Y, M, D);
  int    d = int (D);
  double H = 24.* (D - double (d));
  int    h = int (H);
  double N = 60. * (H - double (h));
  int    n = int (N);
  double S = 60. * (N - double (n));
  int    s = rint (S);
  if (s == 60)
    {
      n += 1;
      s  = 0;
    }

  double JDU = JD - DeltaT (JD) /day;
  int YU, MU;
  double DU;
  JuliantoDate (JDU, YU, MU, DU);
  int    dU = int (DU);
  double HU = 24.* (DU - double (dU));
  int    hU = int (HU);
  double NU = 60. * (HU - double (hU));
  int    nU = int (NU);
  double SU = 60. * (NU - double (nU));
  int    sU = rint (SU);
  if (sU == 60)
    {
      nU += 1;
      sU  = 0;
    }

  printf ("Spring Equinox : March %2d,     %02d:%02d:%02d TDT,   %02d:%02d:%02d UT\n",
	  d, h, n, s, hU, nU, sU);
 
  // Find summer solstice
  M = 6; D = 21.;
  DatetoJulian (Y, M, D, JD);
  do
    {
      // Get geocentric ecliptic coordinates of Sun
      double r, beta;
      SunEcliptic (JD, r, beta, lambda, 1, 1, 1, 1);
      
      JD -= (lambda - M_PI/2.) * (365.24/2./M_PI);
    }
  while (fabs (lambda - M_PI/2.) > 1.e-10);

  JuliantoDate (JD, Y, M, D);
  d = int (D);
  H = 24.* (D - double (d));
  h = int (H);
  N = 60. * (H - double (h));
  n = int (N);
  S = 60. * (N - double (n));
  s = rint (S);
  if (s == 60)
    {
      n += 1;
      s  = 0;
    }

  JDU = JD - DeltaT (JD) /day;
  JuliantoDate (JDU, YU, MU, DU);
  dU = int (DU);
  HU = 24.* (DU - double (dU));
  hU = int (HU);
  NU = 60. * (HU - double (hU));
  nU = int (NU);
  SU = 60. * (NU - double (nU));
  sU = rint (SU);
  if (sU == 60)
    {
      nU += 1;
      sU  = 0;
    }

  printf ("Summer Solstice: June %2d,      %02d:%02d:%02d TDT,   %02d:%02d:%02d UT\n",
	  d, h, n, s, hU, nU, sU);

  // Find autumn equinox
  M = 9; D = 22.;
  DatetoJulian (Y, M, D, JD);
  do
    {
      // Get geocentric ecliptic coordinates of Sun
      double r, beta;
      SunEcliptic (JD, r, beta, lambda, 1, 1, 1, 1);

      if (lambda < 0.) lambda += 2.*M_PI;
      JD -= (lambda - M_PI) * (365.24/2./M_PI);
    }
  while (fabs (lambda - M_PI) > 1.e-10);

  JuliantoDate (JD, Y, M, D);
  d = int (D);
  H = 24.* (D - double (d));
  h = int (H);
  N = 60. * (H - double (h));
  n = int (N);
  S = 60. * (N - double (n));
  s = rint (S);
  if (s == 60)
    {
      n += 1;
      s  = 0;
    }

  JDU = JD - DeltaT (JD) /day;
  JuliantoDate (JDU, YU, MU, DU);
  dU = int (DU);
  HU = 24.* (DU - double (dU));
  hU = int (HU);
  NU = 60. * (HU - double (hU));
  nU = int (NU);
  SU = 60. * (NU - double (nU));
  sU = rint (SU);
  if (sU == 60)
    {
      nU += 1;
      sU  = 0;
    }

  printf ("Autumn Equinox : September %2d, %02d:%02d:%02d TDT,   %02d:%02d:%02d UT\n",
	  d, h, n, s, hU, nU, sU);

  // Find winter solstice
  M = 12; D = 21.;
  DatetoJulian (Y, M, D, JD);
  do
    {
      // Get geocentric ecliptic coordinates of Sun
      double r, beta;
      SunEcliptic (JD, r, beta, lambda, 1, 1, 1, 1);
      
       if (lambda < 0.) lambda += 2.*M_PI;
      JD -= (lambda  - 3.*M_PI/2.) * (365.24/2./M_PI);
    }
  while (fabs (lambda - 3.*M_PI/2.) > 1.e-10);

  JuliantoDate (JD, Y, M, D);
  d = int (D);
  H = 24.* (D - double (d));
  h = int (H);
  N = 60. * (H - double (h));
  n = int (N);
  S = 60. * (N - double (n));
  s = rint (S);
  if (s == 60)
    {
      n += 1;
      s  = 0;
    }

  JDU = JD - DeltaT (JD) /day;
  JuliantoDate (JDU, YU, MU, DU);
  dU = int (DU);
  HU = 24.* (DU - double (dU));
  hU = int (HU);
  NU = 60. * (HU - double (hU));
  nU = int (NU);
  SU = 60. * (NU - double (nU));
  sU = rint (SU);
  if (sU == 60)
    {
      nU += 1;
      sU  = 0;
    }

  printf ("Winter Solstice: December %2d,  %02d:%02d:%02d TDT,   %02d:%02d:%02d UT\n",
	  d, h, n, s, hU, nU, sU);
}

// =============================================================
// Function to determine dates of all full moons in a given year.
// Y is year (e.g., 1963).
// =============================================================
void FullMoon (int Y)
{
  printf ("\n");
  printf ("Dates of full moons:\n\n");
  int Y0 = Y;

  const char *MO[12];
  MO[0] = "Jan"; MO[1] = "Feb"; MO[ 2] = "Mar"; MO[ 3] = "Apr";
  MO[4] = "May"; MO[5] = "Jun"; MO[ 6] = "Jul"; MO[ 7] = "Aug";
  MO[8] = "Sep"; MO[9] = "Oct"; MO[10] = "Nov"; MO[11] = "Dec";

  // Find initial lunar elongation
  int    M = 1;
  double D = 1.;
  double JD;
  DatetoJulian (Y, M, D, JD);
  double E = MoonElongation (JD);
  if (E < 0.)
    E += 2.*M_PI;
  
  // Estimate date of first full moon
  if (E < M_PI)
    JD += (M_PI - E) * (29.5306/(2.*M_PI));
  else
    JD += (3.*M_PI - E) * (29.5306/(2.*M_PI));

  // Find dates of full moons
  for (int i = 0; i < 13; i++)
    {
      do
	{
	  E = MoonElongation (JD);
	  if (E < 0.)
	    E += 2.*M_PI;
	  JD += (M_PI - E) * (29.5306/(2.*M_PI));
	}
      while (fabs (E - M_PI) > 1.e-10);

      JuliantoDate (JD, Y, M, D);
      int    d = int (D);
      double H = 24.* (D - double (d));
      int    h = int (H);
      double N = 60. * (H - double (h));
      int    n = int (N);
      double S = 60. * (N - double (n));
      int    s = rint (S);
      if (s == 60)
	{
	  n += 1;
	  s  = 0;
	}

      double JDU = JD - DeltaT (JD) /day;
      int YU, MU;
      double DU;
      JuliantoDate (JDU, YU, MU, DU);
      int    dU = int (DU);
      double HU = 24.* (DU - double (dU));
      int    hU = int (HU);
      double NU = 60. * (HU - double (hU));
      int    nU = int (NU);
      double SU = 60. * (NU - double (nU));
      int    sU = rint (SU);
      if (sU == 60)
	{
	  nU += 1;
	  sU  = 0;
	}
      
      if (Y == Y0)
	printf ("%3s  %2d, %02d:%02d:%02d TDT,   %02d:%02d:%02d UT\n",
		MO[M-1], d, h, n, s, hU, nU, sU);

      JD += 29.5306;
    }
  printf ("\n");
}

// ===========================================================
// Function to determine dates of all new moons in given year.
// Y is year (e.g., 1963).
// ===========================================================
void NewMoon (int Y)
{
  printf ("Dates of new moons:\n\n");
  int Y0 = Y;

  const char *MO[12];
  MO[0] = "Jan"; MO[1] = "Feb"; MO[ 2] = "Mar"; MO[ 3] = "Apr";
  MO[4] = "May"; MO[5] = "Jun"; MO[ 6] = "Jul"; MO[ 7] = "Aug";
  MO[8] = "Sep"; MO[9] = "Oct"; MO[10] = "Nov"; MO[11] = "Dec";

  // Find initial lunar elongation
  int    M = 1;
  double D = 1.;
  double JD;
  DatetoJulian (Y, M, D, JD);
  double E = MoonElongation (JD);
  
  // Estimate date of first new moon
  if (E < 0.)
    JD += (-E) * (29.5306/2./M_PI);
  else
    JD += (2.*M_PI - E) * (29.5306/(2.*M_PI));

  // Find new moons
  for (int i = 0; i < 13; i++)
    {
      do
	{
	  E = MoonElongation (JD);
	  JD -= E * (29.5306/(2.*M_PI));
	}
      while (fabs (E) > 1.e-10);

      JuliantoDate (JD, Y, M, D);
      int    d = int (D);
      double H = 24.* (D - double (d));
      int    h = int (H);
      double N = 60. * (H - double (h));
      int    n = int (N);
      double S = 60. * (N - double (n));
      int    s = rint (S);
      if (s == 60)
	{
	  n += 1;
	  s  = 0;
	}

      double JDU = JD - DeltaT (JD) /day;
      int YU, MU;
      double DU;
      JuliantoDate (JDU, YU, MU, DU);
      int    dU = int (DU);
      double HU = 24.* (DU - double (dU));
      int    hU = int (HU);
      double NU = 60. * (HU - double (hU));
      int    nU = int (NU);
      double SU = 60. * (NU - double (nU));
      int    sU = rint (SU);
      if (sU == 60)
	{
	  nU += 1;
	  sU  = 0;
	}

      if (Y == Y0)
	printf ("%3s  %2d, %02d:%02d:%02d TDT,   %02d:%02d:%02d UT\n",
		MO[M-1], d, h, n, s, hU, nU, sU);
 
      JD += 29.5306;
    }
  printf ("\n");
}

// ==========================================================
// Function to return geocentric elongation of Moon from Sun.
// t is Julian day number. 
// ==========================================================
double MoonElongation (double t)
{
  // Get geocentric ecliptic coordinates of Sun
  double Sun_r, Sun_beta, Sun_lambda;
  SunEcliptic (t, Sun_r, Sun_beta, Sun_lambda, 1, 1, 1, 1);

  // Get geocentric ecliptic coordinates of Moon
  double Moon_r, Moon_beta, Moon_lambda;
  MoonEcliptic (t, Moon_r, Moon_beta, Moon_lambda, 1, 1, 1, 1);

  // Return geocentric elongation of Moon from Sun
  double E = Moon_lambda - Sun_lambda;
  return atan2 (sin (E), cos (E));
}

