// ############################################
// Definitions of ecliptic coordinate functions
// ############################################

#include "Ecliptic.h"

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Mercury.
// t is Julian day number. r in AU.
// ================================================================
void MercuryEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Mercury
  double MercuryX = Mercury_X (t) - Earth_X (t);
  double MercuryY = Mercury_Y (t) - Earth_Y (t);
  double MercuryZ = Mercury_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (MercuryX * MercuryX + MercuryY * MercuryY);

  r      = sqrt (MercuryX * MercuryX + MercuryY * MercuryY + MercuryZ * MercuryZ);
  beta   = atan (MercuryZ /rho);
  lambda = atan2 (MercuryY, MercuryX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      MercuryEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Venus.
// t is Julian day number. r in AU.
// ================================================================
void VenusEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Venus
  double VenusX = Venus_X (t) - Earth_X (t);
  double VenusY = Venus_Y (t) - Earth_Y (t);
  double VenusZ = Venus_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (VenusX * VenusX + VenusY * VenusY);

  r      = sqrt (VenusX * VenusX + VenusY * VenusY + VenusZ * VenusZ);
  beta   = atan (VenusZ /rho);
  lambda = atan2 (VenusY, VenusX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      VenusEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// =============================================================
// Function to calculate geocentric ecliptic coordinates of Sun.
// t is Julian day number. r in AU.
// =============================================================
void SunEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Sun 
  double SunX = - Earth_X (t);
  double SunY = - Earth_Y (t);
  double SunZ = - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (SunX * SunX + SunY * SunY);

  r      = sqrt (SunX * SunX + SunY * SunY + SunZ * SunZ);
  beta   = atan (SunZ /rho);
  lambda = atan2 (SunY, SunX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      SunEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Mars.
// t is Julian day number. r in AU.
// ================================================================
void MarsEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Mars
  double MarsX = Mars_X (t) - Earth_X (t);
  double MarsY = Mars_Y (t) - Earth_Y (t);
  double MarsZ = Mars_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (MarsX * MarsX + MarsY * MarsY);

  r      = sqrt (MarsX * MarsX + MarsY * MarsY + MarsZ * MarsZ);
  beta   = atan (MarsZ /rho);
  lambda = atan2 (MarsY, MarsX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      MarsEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Jupiter.
// t is Julian day number. r in AU.
// ================================================================
void JupiterEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Jupiter
  double JupiterX = Jupiter_X (t) - Earth_X (t);
  double JupiterY = Jupiter_Y (t) - Earth_Y (t);
  double JupiterZ = Jupiter_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (JupiterX * JupiterX + JupiterY * JupiterY);

  r      = sqrt (JupiterX * JupiterX + JupiterY * JupiterY + JupiterZ * JupiterZ);
  beta   = atan (JupiterZ /rho);
  lambda = atan2 (JupiterY, JupiterX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      JupiterEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Saturn.
// t is Julian day number. r in AU.
// ================================================================
void SaturnEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Saturn
  double SaturnX = Saturn_X (t) - Earth_X (t);
  double SaturnY = Saturn_Y (t) - Earth_Y (t);
  double SaturnZ = Saturn_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (SaturnX * SaturnX + SaturnY * SaturnY);

  r      = sqrt (SaturnX * SaturnX + SaturnY * SaturnY + SaturnZ * SaturnZ);
  beta   = atan (SaturnZ /rho);
  lambda = atan2 (SaturnY, SaturnX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      SaturnEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Uranus.
// t is Julian day number. r in AU.
// ================================================================
void UranusEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Uranus
  double UranusX = Uranus_X (t) - Earth_X (t);
  double UranusY = Uranus_Y (t) - Earth_Y (t);
  double UranusZ = Uranus_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (UranusX * UranusX + UranusY * UranusY);

  r      = sqrt (UranusX * UranusX + UranusY * UranusY + UranusZ * UranusZ);
  beta   = atan (UranusZ /rho);
  lambda = atan2 (UranusY, UranusX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      UranusEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// ================================================================
// Function to calculate geocentric ecliptic coordinates of Uranus.
// t is Julian day number. r in AU.
// ================================================================
void NeptuneEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Neptune
  double NeptuneX = Neptune_X (t) - Earth_X (t);
  double NeptuneY = Neptune_Y (t) - Earth_Y (t);
  double NeptuneZ = Neptune_Z (t) - Earth_Z (t);

  // Calculate ecliptic coordinates
  double rho = sqrt (NeptuneX * NeptuneX + NeptuneY * NeptuneY);

  r      = sqrt (NeptuneX * NeptuneX + NeptuneY * NeptuneY + NeptuneZ * NeptuneZ);
  beta   = atan (NeptuneZ /rho);
  lambda = atan2 (NeptuneY, NeptuneX);

  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);

  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      NeptuneEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}

// =============================================================
// Function to calculate geocentric ecliptic coordinates of Moon.
// t is Julian day number. r in AU.
// =============================================================
void MoonEcliptic (double t, double& r, double& beta, double& lambda, int FK5, int Precess, int Nutate, int Aberrate)
{
  // Get geocentric Cartesian coordinates of Moon
  double Moon_X, Moon_Y, Moon_Z;
  Moon (t, Moon_X, Moon_Y, Moon_Z);

  // Calculate ecliptic coordinates
  double rho = sqrt (Moon_X * Moon_X + Moon_Y * Moon_Y);

  r      = sqrt (Moon_X * Moon_X + Moon_Y * Moon_Y + Moon_Z * Moon_Z);
  beta   = atan (Moon_Z /rho);
  lambda = atan2 (Moon_Y, Moon_X);

  // Transform to FK5 system
  if (FK5)
    FK5Correction (t, lambda, beta);
  
  // Correct for precession
  if (Precess)
    Precession (t, beta, lambda);
   
  // Correct for nutation
  if (Nutate)
    {
      double Dpsi;
      NutationLongitude (t, Dpsi);
      lambda += Dpsi;
    }
  
  // Correct for aberration
  if (Aberrate)
    {
      double tau = Abr * r;
      Aberrate = 0;
      MoonEcliptic (t - tau, r, beta, lambda, FK5, Precess, Nutate, Aberrate);
    }

  lambda = atan2 (sin (lambda), cos (lambda));
}
