// #############################
// Definitions of time functions
// #############################

#include "Time.h"

// =================================================================
// Function to convert Julian day number to terrestrial time (TT)
// interval from J2000.0 epoch measured in Julian millenia
// =================================================================
double JD_to_1000 (double JD)
{
  return (JD - J2000) /Jmill;
}

// =================================================================
// Function to convert Julian day number to terrestrial time (TT)
// interval from J2000.0 epoch measured in Julian centuries
// =================================================================
double JD_to_100 (double JD)
{
  return (JD - J2000) /Jcent;
}

// ===================================================================
// Function to convert Julian day number into date.
//
// Date is y/m/d.
//
// y is year number (years counted astronomically, so -584 is 585 BC).
// m is month number (1-12).
// d is day number (with decimal).
//
// JD is Julian day number (with decimal).
// Algorithm fails for negative Julian day number.
// ===================================================================
void JuliantoDate (double JD, int& y, int& m, double& d)
{
  JD      += 0.5;
  int    Z = int (JD);
  double F = JD - double (Z);

  int A;
  if (Z < 2299161)
    A = Z;
  else
    {
      int alpha = int ((Z - 1867216.25) / 36524.25);
      A = Z + 1 + alpha - int (alpha/4);
    }

  int B = A + 1524;
  int C = int ((B - 122.1) / 365.25);
  int D = int (365.25 * C);
  int E = int ((B - D) / 30.6001);

  d = B - D - int (30.6001 * E) + F;
  if (E < 14)
    m = E - 1;
  else
    m = E - 13;
  if (m > 2)
    y = C - 4716;
  else
    y = C - 4715;
}

// ===================================================================
// Function to convert date into Julian day number.
//
// Date is Y/M/D.
// Gregorian calendar assumed after 14/10/1582.
//
// Y is year number (years counted astronomically, so -584 is 585 BC).
// M is month number (1-12).
// D is day number (with decimal).
//
// JD is Julian day number (with decimal).
// Algorithm fails for negative Julian day number.
// ===================================================================
void DatetoJulian (int Y, int M, double D, double& JD)
{
  if (M < 3)
    {
      Y -= 1;
      M += 12;
    }

  int B = 0;
  if (Y + M + int (D) > 1582+10+14)
    {
      int A = int (Y/100);
      B = 2 - A + int (A/4);
    }

  JD = int (365.25 * (Y + 4716)) + int (30.6001 * (M + 1)) + D + B - 1524.5;
}

// ================================================================================================
// Function to calculate difference between terrestrial dynamic time and universal time in seconds.
//
// t is Julian day number.
//
// See http://eclipse.gsfc.nasa.gov/SEhelp/delta2004.html
// ================================================================================================
double DeltaT (double t)
{
  // Get year number
  int Y, M;
  double D, Dt;
  JuliantoDate (t, Y, M, D);
  double y = double (Y) + (double (M) - 0.5) /12.;

  // Calculate Dt
  if (y < - 500)
    {
      double u = (y - 1820.) /100.;
      Dt = - 20. + 32. * u*u; 
    }
  else if (y >= -500. && y < +500.)
    {
      double u = y /100.;
      Dt = 10583.6 - 1014.41 * u + 33.78311 * u*u - 5.952053 * u*u*u
	- 0.1798452 * u*u*u*u + 0.022174192 * u*u*u*u*u + 0.0090316521 * u*u*u*u*u*u;
    }
  else if (y >= +500. && y < +1600.)
    {
      double u = (y - 1000.) /100.;
      Dt = 1574.2 - 556.01 * u + 71.23472 * u*u + 0.319781 * u*u*u - 0.8403463 * u*u*u*u
	- 0.005050998 * u*u*u*u*u + 0.0083572073 * u*u*u*u*u*u;
    }
  else if (y >= +1600. && y < +1700.)
    {
      double u = y - 1600.;
      Dt = 120. - 0.9808 * u - 0.01532 * u*u + u*u*u /7192.;
    }
  else if (y >= +1700. && y < +1800.)
    {
      double u = y - 1700.;
      Dt = 8.83 + 0.1603 * u - 0.0059285 * u*u + 0.000013336 * u*u*u - u*u*u*u /117400.;
    }
  else if (y >= +1800. && y < +1860.)
    {
      double u = y - 1800.;
      Dt = 13.72 - 0.332447 * u + 0.0068612 * u*u + 0.0041116 * u*u*u - 0.00037436 * u*u*u*u
	+ 0.0000121272 * u*u*u*u*u - 0.0000001699 * u*u*u*u*u*u + 0.000000000875 * u*u*u*u*u*u*u;
    }
  else if (y >= +1860. && y < +1900.)
    {
      double u = y - 1860.;
      Dt = 7.62 + 0.5737 * u - 0.251754 * u*u + 0.01680668 * u*u*u - 0.0004473624 * u*u*u*u + u*u*u*u*u /233174.;
    }
  else if (y >= +1900. && y < +1920.)
    {
      double u = y - 1900.;
      Dt = - 2.79 + 1.494119 * u - 0.0598939 * u*u + 0.0061966 * u*u*u - 0.000197 * u*u*u*u;
    }
  else if (y >= +1920. && y < +1941.)
    {
      double u = y - 1920.;
      Dt = 21.20 + 0.84483 * u - 0.076100 * u*u + 0.0020936 * u*u*u;
    }
  else if (y >= +1941. && y < +1961.)
    {
      double u = y - 1950.;
      Dt = 29.07 + 0.407 * u - u*u /233. + u*u*u /2547.;
    }
  else if (y >= +1961. && y < 1986.)
    {
      double u = y - 1975.;
      Dt = 45.45 + 1.067 * u - u*u /260. - u*u*u /718.;
    }
  else if (y >= +1986. && y < +2005.)
    {
      double u = y - 2000.;
      Dt = 63.86 + 0.3345 * u - 0.060374 * u*u + 0.0017275 * u*u*u + 0.000651814 * u*u*u*u
	+ 0.00002373599 * u*u*u*u*u;
    }
  else if (y >= +2005. && y < +2050.)
    {
      double u = y - 2000.;
      Dt = 62.92 + 0.32217 * u + 0.005589 * u*u;
    }
  else if (y >= +2050. && y < 2150.)
    {
      double u = (y - 1820.) /100.;
      Dt = - 20. + 32. * u*u - 0.5628 * (2150. - y);
    }
  else
    {
      double u = (y - 1820.) /100.;
      Dt = - 20. + 32 * u*u;
    }

  if (y < +1955 || y > +2005)
    {
      double u = (y - 1955.);
      Dt -= 0.000012932 * u*u;
    }

  return Dt;
}

// ===========================================================
// Calculate mean Greewich sidereal time (in radians) at date.
// t is Julian day number
// ===========================================================
double MeanGreenwichSiderealTime (double t)
{
  // Calculate time interval in UT since J2000
  double dt = t - J2000 - (DeltaT (t) - DeltaT (J2000)) /day;

  // Calculate theta
  double theta = (M_PI/180.) * (280.46061837 + 360.98564736629 * dt);

  return theta;
}

// ===============================================================
// Convert real day number within month into integer day and time.
//
// D - real day number within month
// d - integer day number within month
// h - integer hour
// m - integer minute
// s - integer second (nearest)
//
// ===============================================================
void DayNumberConvert (double D, int& d, int& h, int& m, int& s)
{
         d = int (D);
  double H = 24.* (D - double (d));
         h = int (H);
  double M = 60. * (H - double (h));
         m = int (M);
  double S = 60. * (M - double (m));
         s = rint (S);
  if (s == 60)
    {
      m += 1;
      s  = 0;
    }
}

// ============================================================
// Convert real day number within month into reduced day number
// corresponding to previous integer half hour.
//
// D - real day number within month
// E - reduced day number
//
// ============================================================
void DayNumberReduce (double D, double &E)
{
  int d    = int (D);
  double H = 24.* (D - double (d));
  int h    = int (H);
  double M = 60. * (H - double (h));
  int m    = int (M);

  E = double (d) + double (h) /24.;
  if (m > 29) E += 1./48;
}

// ==============================================================
// Convert real time in seconds into integer minutes and seconds.
//
// t - real time in seconds
// m - integer minute
// s - integer second (nearest)
//
// ==============================================================
void TimeConvert (double t, int& m, int& s)
{
         m = int (t /60.);
  double S = t - 60. * double (m);
         s = rint (S);
  if (s == 60)
    {
      m += 1;
      s  = 0;
    }
}

// ===============================================================================
// Convert real declination in radians into integer degrees, minutes, and seconds.
//
// t - real declination in radians
// d - integer degree
// m - integer minute
// s - integer second (nearest)
//
// ===============================================================================
void DeclinationConvert (double t, int& d, int& m, int& s)
{
  t *= (180. /M_PI);
  int sign;
  if (t < 0)
    {
      sign = -1;
      t    = -t;
    }
  else
    sign = +1;
  
         d = int (t);
  double M = 60. * (t - double (d));
         m = int (M);
  double S = 60. * (M - double (m));
         s = rint (S);

  if (s == 60)
    {
      m += 1;
      s  = 0;
    }

  d *= sign;
}

// =================================================================================
// Convert real right ascension in radians into integer hours, minutes, and seconds.
//
// t - real right ascension in radians
// h - integer hour
// m - integer minute
// s - integer second (nearest)
//
// =================================================================================
void RightAscensionConvert (double t, int& h, int& m, int& s)
{
  t *= (12. /M_PI);
  if (t < 0.)
    t += 24.;
  
         h = int (t);
  double M = 60. * (t - double (h));
         m = int (M);
  double S = 60. * (M - double (m));
         s = rint (S);

  if (s == 60)
    {
      m += 1;
      s  = 0;
    }
}

